def compute_hexdigest(text):
  from libtbx.utils import hashlib_md5
  m = hashlib_md5()
  m.update(text)
  return m.hexdigest()

def check_fingerprint(file_name):
  from libtbx.utils import hashlib_md5
  lines = open(file_name).read().splitlines()
  if (len(lines) == 0): return None
  flds = lines[0].split()
  if (len(flds) < 2 or flds[-2] != "fingerprint"): return None
  orig_hexdigest = flds[-1]
  curr_text = "\n".join(lines[1:])+"\n"
  curr_hexdigest = compute_hexdigest(text=curr_text)
  if (len(orig_hexdigest) != len(curr_hexdigest)): return None
  return (orig_hexdigest == curr_hexdigest)

def write_only_if_safe(file_name, text):
  from libtbx.str_utils import show_string
  import os
  op = os.path
  if (op.exists(file_name)):
    if (not op.isfile(file_name)):
      raise RuntimeError(
        "Not a regular file: %s" % show_string(file_name))
    stat = check_fingerprint(file_name=file_name)
    if (stat is None or not stat):
      raise RuntimeError(
        "File appears to be manually modified: %s" % show_string(file_name))
  hexdigest = compute_hexdigest(text=text)
  f = open(file_name, "w")
  f.write("// fingerprint %s\n" % hexdigest)
  f.write(text)

class process(object):

  __slots__ = ["options"]

  def __init__(O, options):
    O.options = options

  def __call__(O, file_names):
    import fable.cout
    import sys
    opts = O.options
    lines = fable.cout.process(
      file_names=file_names,
      top_unit_name=opts.top_unit_name,
      debug=opts.debug)
    text = "\n".join(lines)+"\n"
    if (opts.top_unit_name is None or not opts.debug):
      sys.stdout.write(text)
    if (len(file_names) != 0):
      print
    if (len(file_names) != 0 and opts.compile):
      write_only_if_safe(file_name="fable_cout.cpp", text=text)
      from fable import simple_compilation
      comp_env = simple_compilation.environment()
      out_name = comp_env.build(
        link=opts.link, file_name_cpp="fable_cout.cpp", show_command=True)
      print
      if (opts.run):
        from libtbx import easy_run
        import os
        op = os.path
        cmd = op.join(".", out_name)
        if (opts.valgrind):
          cmd = "valgrind " + cmd
        print cmd
        easy_run.call(command=cmd)
        print

def run(args):
  if (len(args) == 0): args = ["--help"]
  from libtbx.option_parser import option_parser
  import libtbx.load_env
  command_line = (option_parser(
    usage="%s [options] fortran_file ..." % libtbx.env.dispatcher_name)
    .option(None, "--compile",
      action="store_true",
      default=False)
    .option(None, "--link",
      action="store_true",
      default=False)
    .option(None, "--run",
      action="store_true",
      default=False)
    .option(None, "--valgrind",
      action="store_true",
      default=False)
    .option(None, "--each",
      action="store_true",
      default=False)
    .option(None, "--top_unit_name",
      action="store",
      type="str")
    .option(None, "--debug",
      action="store_true",
      default=False)
  ).process(args=args)
  co = command_line.options
  if (co.valgrind): co.run = True
  if (co.run): co.link = True
  if (co.link): co.compile = True
  if (not co.each):
    process(options=co)(file_names=command_line.args)
  else:
    from fable.command_line.read import process_each
    process_each(process=process(options=co), file_names=command_line.args)

if (__name__ == "__main__"):
  import sys
  run(args=sys.argv[1:])
