def show_traceback():
  import traceback
  print traceback.format_exc(limit=None)

def strip_leading_zeros(string):
  for i in xrange(len(string)):
    if (string[i] != "0"):
      return string[i:]
  if (len(string) == 0):
    return ""
  return "0"

def set_var_prefix(var_prefixes, fdecl):
  identifier = fdecl.id_tok.value
  if (fdecl.is_common()):
    var_prefixes[identifier] = "cmn."
  elif (fdecl.is_save()):
    var_prefixes[identifier] = "sve."
  elif (fdecl.is_intrinsic()):
    if (identifier in ["float", "int"]):
      var_prefixes[identifier] = "fem::f"
    else:
      var_prefixes[identifier] = "fem::"
  else:
    var_prefixes[identifier] = ""

def raise_unhandled(tok):
  raise RuntimeError("Unhandled token %s %s" % (tok.type(), tok.value))

def convert_token(var_prefixes, leading, tok):
  tv = tok.value
  if (tok.is_identifier()):
    return var_prefixes.get(tv, "") + tv
  if (tok.is_op()):
    if (tv == ".not."):
      return "!"
    if (tv == ".and."):
      return " && "
    if (tv == ".or."):
      return " || "
    if (tv in ["+", "-"]):
      if (leading):
        return tv
      return " "+tv+" "
    if (tv == "*"):
      if (leading):
        return "star "
      return " "+tv+" "
    if (tv == "/"):
      return " "+tv+" "
    if (tv == "//"):
      return " + "
    if (tv == ".eq."):
      return " == "
    if (tv == ".ne."):
      return " != "
    if (tv == ".lt."):
      return " < "
    if (tv == ".le."):
      return " <= "
    if (tv == ".gt."):
      return " > "
    if (tv == ".ge."):
      return " >= "
    if (tv == ":"):
      return ", "
    raise_unhandled(tok=tok)
  if (tok.is_string()):
    return '"'+tv.replace("\\","\\\\").replace('"','\\"')+'"'
  if (tok.is_logical()):
    if (tv == ".false."):
      return "false"
    return "true"
  if (tok.is_integer()):
    return tv
  if (tok.is_real()):
    return tv+"f"
  if (tok.is_double_precision()):
    return tv.replace("d", "e")
  raise_unhandled(tok=tok)

def convert_tokens(var_prefixes, tokens, commas=False):
  result = []
  rapp = result.append
  from tokenization import group_power
  for tok in group_power(tokens=tokens):
    if (tok.is_seq()):
      rapp(convert_tokens(
        var_prefixes=var_prefixes, tokens=tok.value, commas=False))
    elif (tok.is_parentheses()):
      rapp("(" + convert_tokens(
        var_prefixes=var_prefixes, tokens=tok.value, commas=True) + ")")
    elif (tok.is_implied_do()):
      raise AssertionError
    elif (tok.is_power()):
      rapp("fem::pow(" + convert_tokens(
        var_prefixes=var_prefixes, tokens=tok.value, commas=True) + ")")
    else:
      rapp(convert_token(
        var_prefixes=var_prefixes,
        leading=(len(result)==0),
        tok=tok))
  if (commas):
    return ", ".join(result)
  return "".join(result)

def convert_to_int_literal(tokens):
  assert tokens is not None and len(tokens) != 0
  if (len(tokens) != 1 or not tokens[0].is_integer()):
    tokens[0].raise_not_supported()
  return int(strip_leading_zeros(string=tokens[0].value))

def convert_data_type(unit, fdecl, crhs):
  if (fdecl.data_type is None):
    assert unit is not None
    fdecl.data_type = unit.implicit.get(fdecl.id_tok.value[0])
    if (fdecl.data_type is None):
      raise fdecl.id_tok.raise_semantic_error(msg="Missing data type")
  if (isinstance(fdecl.data_type, str)):
    data_type_code = fdecl.data_type
  else:
    data_type_code = fdecl.data_type.value
  size_tokens = fdecl.size_tokens
  dim_tokens = fdecl.dim_tokens
  if (data_type_code == "character"):
    if (size_tokens is None):
      ctype = "char";
      if (crhs is not None and crhs.startswith('"')):
        crhs = "'%s'" % crhs[1:2]
    else:
      csize = convert_tokens(var_prefixes={}, tokens=size_tokens)
      ctype = "fem::str<%s>" % csize
    if (crhs is None):
      crhs = "fem::char0";
  else:
    def convert_to_ctype_with_size(ctype):
      if (size_tokens is None):
        return ctype
      return "fem::%s_star_%d" % (data_type_code, convert_to_int_literal(
        tokens=size_tokens))
    if (data_type_code == "logical"):
      ctype = convert_to_ctype_with_size(ctype="bool")
    elif (data_type_code == "integer"):
      ctype = convert_to_ctype_with_size(ctype="int")
    elif (data_type_code == "real"):
      ctype = convert_to_ctype_with_size(ctype="float")
    elif (data_type_code == "doubleprecision"):
      if (size_tokens is not None):
        size_tokens[0].raise_syntax_error()
      ctype = "double"
    elif (data_type_code == "byte"):
      if (size_tokens is not None):
        size_tokens[0].raise_syntax_error()
      ctype = "char"
    elif (data_type_code == "complex"):
      if (size_tokens is not None):
        size_tokens[0].raise_not_supported()
      ctype = "std::complex<float>"
    else:
      raise RuntimeError(
        "Not implemented: data_type_code = %s" % data_type_code)
  return ctype, crhs

def convert_dims(var_prefixes, dim_tokens):
  need_origin = False
  for tokens in dim_tokens:
    for tok in tokens.value:
      if (tok.is_op_with(value=":")):
        need_origin = True
        break
  dims = []
  for i_dim,tokens in enumerate(dim_tokens):
    cdim = convert_tokens(var_prefixes=var_prefixes, tokens=tokens.value)
    if (cdim == "star "):
      cdim = "star"
    else:
      cdim = cdim.replace(",  * ", ", star") # XXX
    if (need_origin):
      dims.append("dim%d(%s)" % (i_dim+1, cdim))
    else:
      dims.append(cdim)
  if (need_origin):
    result = ".".join(dims)
  else:
    result = "dimension(" + ", ".join(dims) + ")"
  return result

def convert_data_type_and_dims(unit, fdecl, crhs, var_prefixes={}):
  ctype, crhs = convert_data_type(unit=unit, fdecl=fdecl, crhs=crhs)
  if (fdecl.dim_tokens is None):
    cdims = None
  else:
    cdims = convert_dims(
      var_prefixes=var_prefixes, dim_tokens=fdecl.dim_tokens)
  return ctype, cdims, crhs

def zero_shortcut_if_possible(ctype):
  if (ctype.startswith("fem::")):
    if (ctype.endswith(">")): s = " "
    else:                     s = ""
    return "fem::zero<%s%s>()" % (ctype, s)
  return "fem::%s0" % ctype

def convert_declaration(rapp, unit, fdecl, crhs, const, var_prefixes={}):
  ctype, cdims, crhs = convert_data_type_and_dims(
    unit=unit, fdecl=fdecl, crhs=crhs, var_prefixes=var_prefixes)
  if (cdims is None):
    if (crhs is None): crhs = zero_shortcut_if_possible(ctype=ctype)
    def static_const():
      if (const): return "static const "
      return ""
    rapp("%s%s %s = %s;" % (static_const(), ctype, fdecl.id_tok.value, crhs))
    return False
  if (len(fdecl.dim_tokens) != 1):
    t = "%s, %d" % (ctype, len(fdecl.dim_tokens))
  else:
    t = ctype
  if (t.endswith(">")): templs = " "
  else:                 templs = ""
  rapp("array<%s%s> %s(%s);" % (t, templs, fdecl.id_tok.value, cdims))
  rapp("%s.fill(%s);" % (
    fdecl.id_tok.value, zero_shortcut_if_possible(ctype=ctype)))
  return True

class scope(object):

  __slots__ = [
    "parent",
    "opening_text",
    "closing_text",
    "data",
    "insert_point",
    "last_is_statement_label",
    "tail"]

  def __init__(O, parent, opening_text=None):
    O.parent = parent
    O.opening_text = opening_text
    O.closing_text = None
    O.data = []
    O.insert_point = None
    O.last_is_statement_label = False
    O.tail = None

  def remember_insert_point(O):
    O.insert_point = len(O.data)

  def top_append(O, obj):
    assert O.insert_point is not None
    O.data.insert(O.insert_point, obj)
    O.insert_point += 1

  def append(O, obj):
    O.data.append(obj)
    O.last_is_statement_label = False

  def append_statement_label(O, label):
    O.data.append("statement_%s:" % label)
    O.last_is_statement_label = True

  def open_nested_scope(O, opening_text):
    O.last_is_statement_label = False
    return scope(parent=O, opening_text=opening_text)

  def finalize(O):
    if (O.last_is_statement_label):
      O.data[-1] += ";"

  def close_nested_scope(O):
    assert O.opening_text is not None
    assert O.closing_text is None
    assert O.tail is None
    O.finalize()
    O.closing_text = ["}"]
    head = O
    while (head.parent.tail is head):
      head = head.parent
    head.parent.data.append(head)
    return head.parent

  def attach_tail(O, opening_text):
    assert O.opening_text is not None
    assert O.closing_text is None
    assert O.tail is None
    O.closing_text = ["}"]
    O.tail = scope(parent=O, opening_text=opening_text)
    return O.tail

  def collect_text(O, callback, indent="  "):
    for obj in O.data:
      if (isinstance(obj, scope)):
        curr = obj
        while (curr is not None):
          for text in curr.opening_text:
            callback(indent+text)
          curr.collect_text(callback=callback, indent=indent+"  ")
          for text in curr.closing_text:
            callback(indent+text)
          curr = curr.tail
      else:
        callback(indent+obj)

def convert_inquire_and_open(
      var_prefixes,
      curr_scope,
      io_function,
      io_function_specialization,
      io_call_args,
      iolist):
  if (iolist.err is None):
    io_scope = curr_scope
  else:
    io_scope = curr_scope.open_nested_scope(opening_text=["try {"])
  io_scope.append("cmn.io.%s%s(%s)" % (
    io_function, io_function_specialization, io_call_args))
  for slot in iolist.chain:
    tokens = getattr(iolist, slot)
    if (tokens is not None):
      carg = convert_tokens(var_prefixes=var_prefixes, tokens=tokens)
      io_scope.append("  .%s(%s)" % (slot, carg))
  io_scope.data[-1] += ";"
  if (io_scope is not curr_scope):
    io_scope.close_nested_scope()
    catch_scope = curr_scope.open_nested_scope(
      opening_text=["catch (fem::%s_err const&) {" % io_function])
    clabel = convert_tokens(var_prefixes=var_prefixes, tokens=iolist.err)
    catch_scope.append("goto statement_%s;" % clabel)
    catch_scope.close_nested_scope()

def find_implied_dos(result, tokens):
  assert isinstance(tokens, list)
  for i,tok in enumerate(tokens):
    if (tok.is_seq_or_parentheses()):
      find_implied_dos(result=result, tokens=tok.value)
    elif (tok.is_implied_do()):
      result.append(tok)

def convert_io_loop(io_scope, io_op, var_prefixes, tokens, cbuf=None):
  class cbuffer(object):
    __slots__ = ["strings", "leading"]
    def __init__(O):
      O.strings = []
      O.leading = True
    def append(O, string):
      O.strings.append(string)
      O.leading = False
    def append_comma(O):
      if (len(O.strings) != 0 and O.strings[-1] != ", "):
        O.strings.append(", ")
      O.leading = True
    def remove_trailing_comma(O):
      if (len(O.strings) != 0):
        assert O.strings[-1] == ", "
        O.strings.pop()
        assert O.leading
        O.leading = False
    def append_opening_parenthesis(O):
      O.strings.append("(")
      O.leading = True
    def append_closing_parenthesis(O):
      O.remove_trailing_comma()
      O.strings.append(")")
    def flush(O):
      O.remove_trailing_comma()
      if (len(O.strings) != 0):
        io_scope.append("%s, %s;" % (io_op, "".join(O.strings)))
        O.strings = []
  if (cbuf is None):
    cbuf = cbuffer()
    owning_cbuf = True
  else:
    owning_cbuf = False
  for tok in tokens:
    if (tok.is_seq()):
      convert_io_loop(
        io_scope, io_op, var_prefixes, tokens=tok.value, cbuf=cbuf)
      cbuf.append_comma()
    elif (tok.is_parentheses()):
      cbuf.append_opening_parenthesis()
      convert_io_loop(
        io_scope, io_op, var_prefixes, tokens=tok.value, cbuf=cbuf)
      cbuf.append_closing_parenthesis()
    elif (tok.is_implied_do()):
      cbuf.flush()
      from fable.tokenization import implied_do_info
      idi = implied_do_info(tokens=tok.value)
      i = convert_token(var_prefixes=var_prefixes,leading=True, tok=idi.id_tok)
      f = convert_tokens(var_prefixes=var_prefixes, tokens=idi.first_tokens)
      l = convert_tokens(var_prefixes=var_prefixes, tokens=idi.last_tokens)
      if (idi.step_tokens is None):
        opening_text = ["FEM_DO(%s, %s, %s) {" % (i, f, l)]
      else:
        s = convert_tokens(var_prefixes=var_prefixes, tokens=idi.step_tokens)
        opening_text = ["FEM_DOSTEP(%s, %s, %s, %s) {" % (i, f, l, s)]
      do_scope = io_scope.open_nested_scope(opening_text=opening_text)
      convert_io_loop(
        io_scope=do_scope,
        io_op=io_op,
        var_prefixes=var_prefixes,
        tokens=tok.value[:idi.dlist_size])
      do_scope.close_nested_scope()
      return
    else:
      cbuf.append(convert_token(
        var_prefixes=var_prefixes,
        leading=cbuf.leading,
        tok=tok))
  if (owning_cbuf):
    cbuf.flush()

def convert_data(unit, var_prefixes, data_init_scope):
  for nlist,clist in unit.data:
    ccs = []
    for repetition_tok,ctoks in clist:
      cc = convert_tokens(var_prefixes=var_prefixes, tokens=ctoks)
      if (repetition_tok is not None):
        cr = convert_tokens(
          var_prefixes=var_prefixes, tokens=[repetition_tok])
        cc = "%s*datum(%s)" % (cr, cc)
      ccs.append(cc)
    implied_dos = []
    find_implied_dos(result=implied_dos, tokens=nlist)
    if (len(implied_dos) == 0):
      cn = convert_tokens(var_prefixes=var_prefixes, tokens=nlist, commas=True)
      data_init_scope.append("data((values, %s)), %s;" % (", ".join(ccs), cn))
    else:
      data_scope = data_init_scope.open_nested_scope(opening_text=["{"])
      data_scope.append("data_values data((values, %s));" % ", ".join(ccs))
      convert_io_loop(
        io_scope=data_scope,
        io_op="data",
        var_prefixes=var_prefixes,
        tokens=nlist)
      data_scope.close_nested_scope()

def declare_identifier(
      unit, var_prefixes, top_scope, curr_scope, id_tok, crhs=None):
  fdecl = unit.get_fdecl(id_tok=id_tok)
  set_var_prefix(var_prefixes, fdecl)
  if (fdecl is not None
        and (   fdecl.is_local()
             or fdecl.is_parameter())):
    const = False
    have_crhs = (crhs is not None)
    have_goto = (len(unit.target_statement_labels()) != 0)
    if (have_goto or curr_scope != top_scope):
      crhs = None
    if (crhs is None):
      if (fdecl.parameter_assignment_tokens is not None):
        crhs = convert_tokens(
          var_prefixes=var_prefixes,
          tokens=fdecl.parameter_assignment_tokens)
        const = True
    elif (fdecl.parameter_assignment_tokens is not None):
      id_tok.raise_semantic_error(
        msg="Assignment to PARAMETER %s" % id_tok.value)
    if (have_goto):
      rapp = top_scope.top_append
    else:
      rapp = top_scope.append
    for sd_tokens in [fdecl.size_tokens, fdecl.dim_tokens]:
      if (sd_tokens is None):
        continue
      from fable.tokenization import extract_identifiers
      sd_id_tokens = extract_identifiers(tokens=sd_tokens)
      for sd_id_tok in sd_id_tokens:
        if (sd_id_tok.value == id_tok.value):
          sd_id_tok.raise_semantic_error(msg="Recursion in declaration")
        if (sd_id_tok.value in var_prefixes):
          continue
        sd_fdecl = unit.get_fdecl(id_tok=sd_id_tok)
        if (sd_fdecl.parameter_assignment_tokens is None):
          sd_crhs = None
        else:
          sd_crhs = convert_tokens(
            var_prefixes=var_prefixes,
            tokens=sd_fdecl.parameter_assignment_tokens)
        set_var_prefix(var_prefixes, sd_fdecl)
        convert_declaration(
          rapp=rapp,
          unit=unit,
          fdecl=sd_fdecl,
          crhs=sd_crhs,
          const=True,
          var_prefixes=var_prefixes)
    result = convert_declaration(
      rapp=rapp,
      unit=unit,
      fdecl=fdecl,
      crhs=crhs,
      const=const,
      var_prefixes=var_prefixes)
    if (have_crhs and (have_goto or curr_scope != top_scope)):
      result = True
    return result
  elif (crhs is not None):
    return True
  return False

def convert_executable(callback, units_by_name, unit, var_prefixes):
  top_scope = scope(parent=None)
  if (unit.uses_read):
    top_scope.append("common_read read(cmn);")
  if (unit.uses_write):
    top_scope.append("common_write write(cmn);")
  top_scope.remember_insert_point()
  def declare_identifiers(id_tokens):
    for id_tok in id_tokens:
      if (id_tok.value not in var_prefixes):
        declare_identifier(
          unit=unit,
          var_prefixes=var_prefixes,
          top_scope=top_scope,
          curr_scope=curr_scope,
          id_tok=id_tok)
  from fable.tokenization import extract_identifiers
  curr_scope = top_scope
  if (len(unit.data) != 0):
    data_init_scope = top_scope.open_nested_scope(
      opening_text=["if (sve.data_init()) {"])
    for nlist,clist in unit.data:
      declare_identifiers(
        id_tokens=extract_identifiers(tokens=nlist))
      for repetition_tok,ctoks in clist:
        if (repetition_tok is not None):
          declare_identifiers(
            id_tokens=extract_identifiers(tokens=[repetition_tok]))
        declare_identifiers(
          id_tokens=extract_identifiers(tokens=ctoks))
    convert_data(
      unit=unit, var_prefixes=var_prefixes, data_init_scope=data_init_scope)
    data_init_scope.close_nested_scope()
  def curr_scope_append_return_function():
    n = unit.name.value
    curr_scope.append("return %s%s;" % (var_prefixes[n], n))
  close_scope_after_next_executable = False
  dos_to_close_by_label = {}
  from fable.tokenization import fmt_tokens_as_string
  from fable.read import Error
  from fable import SemanticError
  for ei in unit.executable:
    if (    ei.ssl.label is not None
        and ei.ssl.label in unit.target_statement_labels()):
      curr_scope.append_statement_label(label=ei.ssl.label)
    def search_for_id_tokens_and_declare_identifiers():
      id_tokens = []
      def callback(tok, next_tok):
        id_tokens.append(tok)
      ei.search_for_id_tokens(callback=callback)
      declare_identifiers(id_tokens=id_tokens)
      return id_tokens
    try:
      if (ei.key == "assignment"):
        lhs_id_tokens = extract_identifiers(tokens=ei.lhs_tokens)
        assert len(lhs_id_tokens) != 0
        rhs_id_tokens = extract_identifiers(tokens=ei.rhs_tokens)
        for id_tokens in lhs_id_tokens[1:], rhs_id_tokens:
          declare_identifiers(id_tokens=id_tokens)
        crhs = convert_tokens(
          var_prefixes=var_prefixes, tokens=ei.rhs_tokens)
        id_tok = lhs_id_tokens[0]
        assign_here = id_tok.value in var_prefixes
        if (not assign_here):
          assign_here = declare_identifier(
            unit=unit,
            var_prefixes=var_prefixes,
            top_scope=top_scope,
            curr_scope=curr_scope,
            id_tok=id_tok,
            crhs=crhs)
        clhs = convert_tokens(var_prefixes=var_prefixes, tokens=ei.lhs_tokens)
        if (assign_here):
          curr_scope.append("%s = %s;" % (clhs, crhs))
      elif (ei.key == "inquire"):
        search_for_id_tokens_and_declare_identifiers()
        iuflist = ei.iuflist
        if (iuflist.unit is not None):
          if (iuflist.file is not None):
            ei.ssl.raise_semantic_error(
              "Conflicting UNIT vs. FILE in INQUIRE statement"
              " (exactly one is needed)", i=ei.start)
          io_function_specialization = "_unit"
          uf_tokens = iuflist.unit
        elif (iuflist.file is not None):
          io_function_specialization = "_file"
          uf_tokens = iuflist.file
        else:
          ei.ssl.raise_semantic_error(
            "Missing UNIT or FILE in INQUIRE statement", i=ei.start)
        io_call_args = convert_tokens(
          var_prefixes=var_prefixes, tokens=uf_tokens)
        convert_inquire_and_open(
          var_prefixes=var_prefixes,
          curr_scope=curr_scope,
          io_function="inquire",
          io_function_specialization=io_function_specialization,
          io_call_args=io_call_args,
          iolist=iuflist)
      elif (ei.key == "open"):
        search_for_id_tokens_and_declare_identifiers()
        olist = ei.olist
        if (olist.unit is None):
          ei.ssl.raise_semantic_error(
            "Missing UNIT in OPEN statement", i=ei.start)
        cunit = convert_tokens(var_prefixes=var_prefixes, tokens=olist.unit)
        if (olist.file is None):
          ei.ssl.raise_semantic_error(
            "Missing FILE in OPEN statement", i=ei.start)
        cfile = convert_tokens(var_prefixes=var_prefixes, tokens=olist.file)
        convert_inquire_and_open(
          var_prefixes=var_prefixes,
          curr_scope=curr_scope,
          io_function="open",
          io_function_specialization="",
          io_call_args="%s, %s" % (cunit, cfile),
          iolist=olist)
      elif (ei.key == "close"):
        search_for_id_tokens_and_declare_identifiers()
        cllist = ei.cllist
        if (cllist.unit is None):
          ei.ssl.raise_semantic_error(
            "Missing UNIT in CLOSE statement", i=ei.start)
        cunit = convert_tokens(var_prefixes=var_prefixes, tokens=cllist.unit)
        if (   cllist.iostat is not None
            or cllist.err is not None
            or cllist.status is not None):
          curr_scope.append("// UNHANDLED cllist")
        curr_scope.append("cmn.io.close(%s);" % cunit)
      elif (ei.key in ["read", "write"]):
        search_for_id_tokens_and_declare_identifiers()
        cilist = ei.cilist
        assert cilist.unit is not None
        cunit = convert_tokens(var_prefixes=var_prefixes, tokens=cilist.unit)
        def conv_fmt():
          tl = cilist.fmt
          if (tl is None): return None
          if (len(tl) != 1): return None
          tok = tl[0]
          if (tok.is_op() and tok.value == "*"):
            return "star"
          if (tok.is_string()):
            return '"' + tok.value.replace('"','\\"') + '"'
          if (tok.is_integer()):
            fmt_tokens = unit.format.get(tok.value)
            if (fmt_tokens is None):
              tok.raise_semantic_error(
                "Unknown FORMAT statement label: %s" % tok.value)
            return '"(' + fmt_tokens_as_string(tokens=fmt_tokens) + ')"'
          return None
        cfmt = conv_fmt()
        if (   cilist.rec is not None
            or cilist.iostat is not None):
          curr_scope.append(
            "// UNHANDLED cilist: %s" % ei.ssl.code_with_strings())
        iolist_id_tokens = extract_identifiers(tokens=ei.iolist)
        declare_identifiers(id_tokens=iolist_id_tokens)
        if (cfmt is None):
          cargs = cunit
        else:
          cargs = "%s, %s" % (cunit, cfmt)
        implied_dos = []
        find_implied_dos(result=implied_dos, tokens=ei.iolist)
        if (len(implied_dos) == 0):
          if (    cilist.end is None
              and cilist.err is None):
            io_scope = curr_scope
          else:
            io_scope = curr_scope.open_nested_scope(opening_text=["try {"])
          io_op = "%s(%s)" % (ei.key, cargs)
          ciolist = convert_tokens(
            var_prefixes=var_prefixes, tokens=ei.iolist, commas=True)
          if (len(ei.iolist) == 0):
            curr_scope.append(io_op+";")
          else:
            convert_io_loop(
              io_scope=io_scope,
              io_op=io_op,
              var_prefixes=var_prefixes,
              tokens=ei.iolist)
        else:
          is_internal_file = False
          unit_id_tokens = extract_identifiers(tokens=cilist.unit)
          if (len(unit_id_tokens) == 1):
            unit_fdecl = unit.get_fdecl(id_tok=unit_id_tokens[0])
            if (    unit_fdecl.data_type is not None
                and unit_fdecl.data_type.value == "character"):
              is_internal_file = True
          if (   cilist.end is not None
              or cilist.err is not None):
            opening_line = "try {"
          else:
            opening_line = "{"
          io_scope = curr_scope.open_nested_scope(opening_text=[opening_line])
          if (is_internal_file): cmn = ""
          else:                  cmn = "cmn, "
          io_scope.append("%s_loop %sloop(%s%s);" % (
            ei.key, ei.key[0], cmn, cargs))
          convert_io_loop(
            io_scope=io_scope,
            io_op=ei.key[0]+"loop",
            var_prefixes=var_prefixes,
            tokens=ei.iolist)
        if (io_scope is not curr_scope):
          io_scope.close_nested_scope()
        for slot in ["end", "err"]:
          tokens = getattr(cilist, slot)
          if (tokens is not None):
            catch_scope = curr_scope.open_nested_scope(
              opening_text=["catch (fem::%s_%s const&) {" % (ei.key, slot)])
            clabel = convert_tokens(var_prefixes={}, tokens=tokens)
            catch_scope.append("goto statement_%s;" % clabel)
            catch_scope.close_nested_scope()
      elif (ei.key == "do"):
        if (ei.id_tok.value not in var_prefixes):
          declare_identifier(
            unit=unit,
            var_prefixes=var_prefixes,
            top_scope=top_scope,
            curr_scope=curr_scope,
            id_tok=ei.id_tok)
        for token in ei.tokens:
          declare_identifiers(
            id_tokens=extract_identifiers(tokens=token.value))
        assert 2 <= len(ei.tokens) <= 3
        i = convert_token(
          var_prefixes=var_prefixes,leading=True, tok=ei.id_tok)
        f = convert_tokens(
          var_prefixes=var_prefixes, tokens=ei.tokens[0].value)
        l = convert_tokens(
          var_prefixes=var_prefixes, tokens=ei.tokens[1].value)
        if (len(ei.tokens) == 3):
          s = convert_tokens(
            tokens=ei.tokens[2].value, var_prefixes=var_prefixes)
          opening_text = ["FEM_DOSTEP(%s, %s, %s, %s) {" % (i, f, l, s)]
        else:
          opening_text = ["FEM_DO(%s, %s, %s) {" % (i, f, l)]
        curr_scope = curr_scope.open_nested_scope(opening_text=opening_text)
        if (ei.label is not None):
          dos_to_close_by_label.setdefault(ei.label, []).append(curr_scope)
      elif (ei.key == "enddo"):
        curr_scope = curr_scope.close_nested_scope()
      elif (ei.key == "if"):
        cond_id_tokens = extract_identifiers(tokens=ei.cond_tokens)
        declare_identifiers(id_tokens=cond_id_tokens)
        c = convert_tokens(var_prefixes=var_prefixes, tokens=ei.cond_tokens)
        curr_scope = curr_scope.open_nested_scope(
          opening_text=["if (%s) {" % c])
        close_scope_after_next_executable = True
        continue
      elif (ei.key == "if_then"):
        cond_id_tokens = extract_identifiers(tokens=ei.cond_tokens)
        declare_identifiers(id_tokens=cond_id_tokens)
        c = convert_tokens(var_prefixes=var_prefixes, tokens=ei.cond_tokens)
        curr_scope = curr_scope.open_nested_scope(
          opening_text=["if (%s) {" % c])
      elif (ei.key == "elseif_then"):
        cond_id_tokens = extract_identifiers(tokens=ei.cond_tokens)
        declare_identifiers(id_tokens=cond_id_tokens)
        c = convert_tokens(var_prefixes=var_prefixes, tokens=ei.cond_tokens)
        curr_scope = curr_scope.attach_tail(
          opening_text=["else if (%s) {" % c])
      elif (ei.key == "else"):
        curr_scope = curr_scope.attach_tail(opening_text=["else {"])
      elif (ei.key == "endif"):
        curr_scope = curr_scope.close_nested_scope()
      elif (ei.key == "call"):
        sub_unit = units_by_name.get(ei.subroutine_name.value)
        if (sub_unit is not None and sub_unit.needs_cmn):
          cmn = "cmn"
        else:
          cmn = ""
        if (ei.arg_token is None):
          curr_scope.append(ei.subroutine_name.value+"(%s);" % cmn)
        else:
          id_tokens = extract_identifiers(tokens=ei.arg_token.value)
          declare_identifiers(id_tokens=id_tokens)
          a = convert_tokens(
            var_prefixes=var_prefixes, tokens=ei.arg_token.value, commas=True)
          def cmn_a():
            if (len(cmn) == 0): return a
            if (len(a) == 0): return cmn
            return cmn + ", " + a
          curr_scope.append("%s(%s);" % (ei.subroutine_name.value, cmn_a()))
      elif (ei.key == "return"):
        if (unit.unit_type == "function"):
          curr_scope_append_return_function()
        elif (ei is not unit.executable[-1]):
          curr_scope.append("return;")
      elif (ei.key == "continue"):
        pass
      elif (ei.key == "goto"):
        curr_scope.append("goto statement_%s;" % ei.label.value)
      elif (ei.key == "stop"):
        if (ei.arg_token is None):
          cmsg = "0"
        elif (ei.arg_token.is_integer()):
          cmsg = strip_leading_zeros(string=ei.arg_token.value)
        else:
          cmsg = convert_token(var_prefixes={}, leading=True, tok=ei.arg_token)
        curr_scope.append("FEM_STOP(%s);" % cmsg)
      else:
        curr_scope.append(
          "// UNHANDLED: executable %s: %s" % (
            ei.key, ei.ssl.code_with_strings()))
      if (close_scope_after_next_executable):
        close_scope_after_next_executable = False
        curr_scope = curr_scope.close_nested_scope()
      if (ei.ssl.label is not None):
        dos_to_close = dos_to_close_by_label.get(ei.ssl.label)
        if (dos_to_close is not None):
          for do_scope in reversed(dos_to_close):
            curr_scope = do_scope.close_nested_scope()
    except (Error, SemanticError):
      raise
    except Exception:
      print "*"*80
      print ei.ssl.format_error(
        i=None,
        msg="Sorry: fable internal error")
      print "*"*80
      print
      raise
  assert curr_scope.parent is None
  if (    unit.unit_type == "function"
      and len(unit.executable) != 0
      and unit.executable[-1].key != "return"):
    curr_scope_append_return_function()
  curr_scope.finalize()
  curr_scope.collect_text(callback=callback)

def convert_to_function_definition(callback, units_by_name, unit):
  var_prefixes = {}
  cargs = []
  if (unit.needs_cmn):
    cargs.append("common& cmn")
  args_dim_buffer = []
  for tok in unit.args:
    if (tok.value == "*"):
      cargs.append("fem::star_type const& /* UNHANDLED: star argument */")
      continue
    assert tok.value not in var_prefixes
    fdecl = unit.get_fdecl(id_tok=tok)
    set_var_prefix(var_prefixes, fdecl)
    assert fdecl.parameter_assignment_tokens is None
    if (fdecl.var_type is None):
      arg_name = "/* %s */" % tok.value
    else:
      arg_name = tok.value
    if (    fdecl.data_type is not None
        and fdecl.data_type.value == "character"
        and fdecl.dim_tokens is None):
      if (fdecl.is_modified): const = ""
      else:                   const = "c"
      cargs.append("fem::str_%sref %s" % (const, arg_name))
    else:
      ctype, crhs = convert_data_type(unit=unit, fdecl=fdecl, crhs=None)
      del crhs
      if (fdecl.dim_tokens is None):
        if (fdecl.is_modified): const = ""
        else:                   const = " const"
        cargs.append("%s%s& %s" % (ctype, const, arg_name))
      else:
        if (fdecl.is_modified): const = ""
        else:                   const = "c"
        if (len(fdecl.dim_tokens) != 1):
          t = "%s, %d" % (ctype, len(fdecl.dim_tokens))
        else:
          t = ctype
        if (t.endswith(">")): templs = " "
        else:                 templs = ""
        cargs.append("fem::%sref<%s%s> %s" % (const, t, templs, arg_name))
        if (fdecl.var_type is not None):
          cdims = convert_dims(
            var_prefixes=var_prefixes, dim_tokens=fdecl.dim_tokens)
          args_dim_buffer.append("  %s(%s);" % (fdecl.id_tok.value, cdims))
  callback("")
  fdecl = unit.get_fdecl(id_tok=unit.name)
  if (fdecl.data_type is None):
    callback("void")
    cdecl = None
  else:
    cdecl, crhs = convert_data_type(unit=unit, fdecl=fdecl, crhs=None)
    assert crhs is None
    del crhs
    callback(cdecl)
    var_prefixes[unit.name.value] = "function_";
  if (len(cargs) == 0):
    callback(unit.name.value+"()")
  else:
    callback(unit.name.value+"(\n  " + ",\n  ".join(cargs) + ")")
  callback("{")
  callback("  using namespace fem::major_types;")
  if (cdecl is not None):
    callback("  %s %s%s;" % (
      cdecl, var_prefixes[unit.name.value], unit.name.value))
  if (unit.uses_save):
    callback("  %s_save& sve = cmn.%s_sve;" % (
      unit.identifier, unit.identifier))
  for ccode in args_dim_buffer:
    callback(ccode)
  convert_executable(
    callback=callback,
    units_by_name=units_by_name,
    unit=unit,
    var_prefixes=var_prefixes)
  callback("}")
  return cdecl, var_prefixes

def convert_to_struct(
      callback, unit, struct_type, struct_name, identifier_list):
  assert struct_type in ["common", "save"]
  callback("")
  callback("struct %s" % struct_name)
  callback("{")
  initializers = []
  if (struct_type == "save" and len(unit.data) != 0):
    callback("  fem::data_init_flag data_init;")
  for id_tok in identifier_list:
    fdecl = unit.get_fdecl(id_tok=id_tok)
    ctype, cdims, crhs = convert_data_type_and_dims(
      unit=unit, fdecl=fdecl, crhs=None)
    if (cdims is None):
      callback("  %s %s;" % (ctype, id_tok.value))
      if (crhs is None):
        crhs = zero_shortcut_if_possible(ctype=ctype)
      initializers.append((id_tok.value, crhs))
    else:
      if (len(fdecl.dim_tokens) != 1):
        t = "%s, %d" % (ctype, len(fdecl.dim_tokens))
      else:
        t = ctype
      if (t.endswith(">")): templs = " "
      else:                 templs = ""
      callback("  fem::array<%s%s> %s;" % (t, templs, id_tok.value))
      initializers.append((id_tok.value, "fem::%s, fem::fill0" % cdims))
  n = len(initializers)
  if (n != 0):
    callback("")
    callback("  %s() :" % struct_name)
    for i in xrange(n):
      if (i+1 == n): comma = ""
      else:          comma = ","
      callback("    %s(%s)" % initializers[i] + comma)
    callback("  {}")
  callback("};")

def convert_commons(callback, units):
  commons_defined_already = set()
  for unit in units:
    for common_name,common_fdecl_list in unit.common.items():
      if (common_name in commons_defined_already): continue
      commons_defined_already.add(common_name)
      identifier_list = []
      for common_fdecl in common_fdecl_list:
        assert common_fdecl.size_tokens is None
        identifier_list.append(common_fdecl.id_tok)
      convert_to_struct(
        callback=callback,
        unit=unit,
        struct_type="common",
        struct_name=common_name,
        identifier_list=identifier_list)
  save_struct_names = []
  for unit in units:
    identifier_list = []
    for fdecl in unit.fdecl_by_identifier.values():
      if (fdecl.is_save()):
        identifier_list.append(fdecl.id_tok)
    if (len(identifier_list) == 0 and len(unit.data) == 0):
      continue
    def id_tok_cmp(a, b):
      return cmp(a.value, b.value)
    identifier_list.sort(id_tok_cmp)
    struct_name = "%s_save" % unit.identifier
    convert_to_struct(
      callback=callback,
      unit=unit,
      struct_type="save",
      struct_name=struct_name,
      identifier_list=identifier_list)
    save_struct_names.append(struct_name)
  if (    len(commons_defined_already) == 0
      and len(save_struct_names) == 0):
    callback("")
    callback("using fem::common;")
    return
  callback("")
  callback("struct common :")
  callback("  " + ",\n  ".join(
    ["fem::common"] + sorted(commons_defined_already)))
  if (len(save_struct_names) == 0):
    callback("{};")
  else:
    callback("{")
    for struct_name in save_struct_names:
      callback("  %s %s;" % (struct_name, struct_name[:-3]+"ve"))
    callback("};")

def convert_program(callback, units, units_by_name, debug):
  main_calls = []
  for unit in units:
    if (not unit.is_program()): continue
    cname = "program_" + unit.identifier
    main_calls.append(cname)
    callback("""
void
%s(
  int argc,
  char const* argv[])
{
  using namespace fem::major_types;
  if (argc != 1) {
    throw std::runtime_error("Unexpected command-line arguments.");
  }""" % cname)
    result_buffer = []
    try:
      convert_executable(
        callback=result_buffer.append,
        units_by_name=units_by_name,
        unit=unit,
        var_prefixes={})
    except Exception:
      if (not debug): raise
      show_traceback()
    else:
      if (unit.needs_cmn):
        callback("  common cmn;")
      if (unit.uses_save):
        callback("  %s_save& sve = cmn.%s_sve;" % (
          unit.identifier, unit.identifier))
      for line in result_buffer:
        callback(line)
      callback("}")
  #
  if (len(main_calls) != 0):
    callback("")
    callback("""\
int
main(
  int argc,
  char const* argv[])
{
  return fem::main_with_catch(argc, argv, %s);
}""" % main_calls[0])

def get_missing_external_return_type(fdecls):
  for fdecl in fdecls:
    if (fdecl.data_type is not None):
      return convert_data_type(unit=None, fdecl=fdecl, crhs=None)[0]
  return "void"

def process(file_names, top_unit_name=None, debug=False):
  import fable.read
  all_units = fable.read.process(file_names=file_names)
  result = []
  if (debug):
    def callback(line):
      print line
      result.append(line)
  else:
    callback = result.append
  callback("""\
#include <fem.hpp> // Fortran EMulation library of fable module""")
  #
  units_by_name = all_units.units_by_name()
  units_bottom_up, missing_external_fdecls_by_identifier = \
    all_units.build_bottom_up_unit_list_following_calls(top_name=top_unit_name)
  if (len(missing_external_fdecls_by_identifier) != 0):
    for identifier in sorted(missing_external_fdecls_by_identifier.keys()):
      return_type = get_missing_external_return_type(
        fdecls=missing_external_fdecls_by_identifier[identifier])
      callback("""
%s
%s(...)
{
  throw std::runtime_error(
    "Missing function implementation: %s");
}""" % (return_type, identifier, identifier))
  #
  try:
    convert_commons(callback=callback, units=units_bottom_up)
  except Exception:
    if (not debug): raise
    show_traceback()
  #
  for unit in units_bottom_up:
    if (   unit.is_function()
        or unit.is_subroutine()):
      try:
        convert_to_function_definition(
          callback=callback, units_by_name=units_by_name, unit=unit)
      except Exception:
        if (not debug): raise
        show_traceback()
  #
  convert_program(
    callback=callback,
    units=units_bottom_up,
    units_by_name=units_by_name,
    debug=debug)
  #
  return result
