#ifndef FEM_MAIN_HPP
#define FEM_MAIN_HPP

#include <fem/stop.hpp>
#include <cstdio>

namespace fem {

  int
  main_with_catch(
    int argc,
    char const* argv[],
    void (*callable)(int argc, char const* argv[]))
  {
    try {
      callable(argc, argv);
    }
    catch (fem::stop_info const& info) {
      std::fflush(stdout);
      std::fprintf(stderr, "%s\n", info.what());
      std::fflush(stderr);
    }
    catch (std::exception const& e) {
      std::fflush(stdout);
      char const* what = e.what();
      if (what == 0) what = "null";
      std::fprintf(stderr, "std::exception what(): %s\n", what);
      std::fflush(stderr);
      return 1;
    }
    catch (...) {
      std::fflush(stdout);
      std::fprintf(stderr, "Terminated by unknown C++ exception.\n");
      std::fflush(stderr);
      return 2;
    }
    return 0;
  }

} // namespace fem

#endif // GUARD
