#ifndef FEM_STR_HPP
#define FEM_STR_HPP

#include <fem/str_ref.hpp>
#include <fem/zero.hpp>

namespace fem {

  template <int StrLen>
  struct str
  {
    char elems[StrLen];

    str()
    {
      std::fill_n(elems, StrLen, '\0');
    }

    str(
      char c)
    {
      std::fill_n(elems, StrLen, c);
    }

    str(
      char const* s)
    {
      utils::copy_with_blank_padding(s, elems, StrLen);
    }

    str(
      str_cref const& other)
    {
      std::memcpy(elems, other.elems(), std::min(StrLen, other.len()));
    }

    str(
      str_cref const& lhs,
      str_cref const& rhs)
    {
      std::memcpy(elems, lhs.elems(), std::min(StrLen, lhs.len()));
      if (lhs.len() < StrLen) {
        std::memcpy(
          elems+lhs.len(), rhs.elems(), std::min(StrLen-lhs.len(), rhs.len()));
      }
    }

    operator
    str_cref() const
    {
      return str_cref(elems, StrLen);
    }

    operator
    str_ref()
    {
      return str_ref(elems, StrLen);
    }

    void
    operator=(
      char const* rhs)
    {
      str_ref(*this) = rhs;
    }

    void
    operator=(
      str_cref const& rhs)
    {
      str_ref(*this) = rhs;
    }

    operator
    std::string() const
    {
      return std::string(elems, StrLen);
    }

    bool
    operator==(
      char const* rhs) const
    {
      return (str_cref(*this) == rhs);
    }

    bool
    operator!=(
      char const* rhs) const
    {
      return (str_cref(*this) != rhs);
    }

    bool
    operator==(
      str_cref const& rhs) const
    {
      return (str_cref(*this) == rhs);
    }

    bool
    operator!=(
      str_cref const& rhs) const
    {
      return (str_cref(*this) != rhs);
    }

    str_cref
    operator()(
      int first,
      int last) const
    {
      return str_cref(*this)(first, last);
    }

    str_ref
    operator()(
      int first,
      int last)
    {
      return str_ref(*this)(first, last);
    }
  };

  template <int StrLen>
  struct zero_impl<str<StrLen> >
  {
    static str<StrLen> get() { return str<StrLen>(); }
  };

  template <int StrLenLhs, int StrLenRhs>
  str<StrLenLhs+StrLenRhs>
  operator+(
    str<StrLenLhs> const& lhs,
    str<StrLenRhs> const& rhs)
  {
    return str<StrLenLhs+StrLenRhs>(lhs, rhs);
  }

} // namespace fem

#endif // GUARD
