#ifndef FEM_STR_REF_HPP
#define FEM_STR_REF_HPP

#include <fem/utils/string.hpp>

namespace fem {

  struct str_cref
  {
    protected:
      char const* elems_;
      int len_;

      public:

    str_cref(
      char const* elems,
      int len)
    :
      elems_(elems),
      len_(len)
    {}

    str_cref(
      char const* elems)
    :
      elems_(elems),
      len_(std::strlen(elems))
    {}

    char const*
    elems() const { return elems_; }

    int
    len() const { return len_; }

    operator
    std::string() const
    {
      return std::string(elems_, len_);
    }

    char const&
    operator[](
      int i) const
    {
      return elems_[i];
    }

    str_cref
    operator()(
      int first,
      int last) const
    {
      return str_cref(elems_+first-1, last-first+1);
    }

    bool
    operator==(
      char const* rhs) const
    {
      return utils::string_eq(elems_, len_, rhs);
    }

    bool
    operator!=(
      char const* rhs) const
    {
      return (!((*this) == rhs));
    }

    bool
    operator==(
      str_cref const& rhs) const
    {
      return utils::string_eq(elems_, len_, rhs.elems_, rhs.len_);
    }

    bool
    operator!=(
      str_cref const& rhs) const
    {
      return (!((*this) == rhs));
    }
  };

  int
  len(
    str_cref const& s) { return s.len(); }

  bool
  operator==(
    char const* lhs,
    str_cref const& rhs) { return (rhs == lhs); }

  bool
  operator!=(
    char const* lhs,
    str_cref const& rhs) { return (rhs != lhs); }

  struct str_ref : str_cref
  {
    str_ref(
      char* elems,
      int len)
    :
      str_cref(elems, len)
    {}

    char*
    elems() const { return const_cast<char*>(elems_); }

    char&
    operator[](
      int i) const
    {
      return elems()[i];
    }

    void
    operator=(
      char const* rhs)
    {
      utils::copy_with_blank_padding(rhs, elems(), len());
    }

    void
    operator=(
      str_cref const& rhs)
    {
      utils::copy_with_blank_padding(rhs.elems(), rhs.len(), elems(), len());
    }

    str_ref
    operator()(
      int first,
      int last) const
    {
      return str_ref(elems()+first-1, last-first+1);
    }
  };

} // namespace fem

#endif // GUARD
