from fable import cout
from libtbx.test_utils import \
  Exception_expected, show_diff, anchored_block_show_diff as absd
import libtbx.load_env
import os
op = os.path

def exercise_simple(verbose):
  t_dir = libtbx.env.under_dist(
    module_name="fable", path="test/valid", test=op.isdir)
  def get(file_name, top_unit_name=None):
    if (verbose):
      print "exercise_simple:", file_name
    file_names = [op.join(t_dir, file_name)]
    lines = cout.process(file_names=file_names, top_unit_name=top_unit_name)
    return "\n".join(lines).splitlines()
  #
  assert not show_diff(get("add_reals.f"), """\
#include <fem.hpp> // Fortran EMulation library of fable module

using fem::common;

void
program_prog(
  int argc,
  char const* argv[])
{
  using namespace fem::major_types;
  if (argc != 1) {
    throw std::runtime_error("Unexpected command-line arguments.");
  }
  float a = fem::float0;
  float b = fem::float0;
  float c = a + b;
}

int
main(
  int argc,
  char const* argv[])
{
  return fem::main_with_catch(argc, argv, program_prog);
}
""")
  #
  assert not absd(get("add_real_integer.f"), -10, """\
  float a = fem::float0;
  int i = fem::int0;
  float c = a + i;
""")
  #
  assert not absd(get("logical_a_or_b.f"), -10, """\
  bool a = fem::bool0;
  bool b = fem::bool0;
  bool c = a || b;
""")
  #
  assert not absd(get("add_dp_integer.f"), -10, """\
  double a = fem::double0;
  int i = fem::int0;
  double c = a + i;
""")
  #
  assert not absd(get("add_strings.f"), -10, """\
  fem::str<3> a = "x\\"z";
  fem::str<4> b = "i\\\\'l";
  fem::str<7> c = a + b;
""")
  #
  assert not absd(get("real_array_sum.f"), -10, """\
  array<float> a(dimension(2));
  a.fill(fem::float0);
  float sum_a = a(1) + a(2);
""")
  #
  assert not absd(get("write_star.f"), -10, """\
  common cmn;
  common_write write(cmn);
  char c = 'x';
  write(6, star), c;
  write(6, star), "i is zero.";
  bool l = fem::bool0;
  write(6, star), l;
  int i = fem::int0;
  write(6, star), i;
  float r = fem::float0;
  write(6, star), r;
  double d = fem::double0;
  write(6, star), d;
  write(6, star), 1.e111;
  write(6, star), -1.e111;
  write(6, star);
  write(6, star), c, c, c, c, c, c, c, c, c, c, c, c;
  write(6, star), "i is ", "zero", ".";
  write(6, star), l, l, l, l, l, l, l, l, l, l, l, l;
  write(6, star), i, i, i, i, i, i, i, i, i, i, i, i;
  write(6, star), r, r, r, r, r, r, r, r, r, r, r, r;
  write(6, star), d, d, d, d, d, d, d, d, d, d, d, d;
  fem::str<1> s1 = "x";
  fem::str<2> s2 = "yz";
  write(6, star), s1, s1;
  write(6, star), s1, s2;
  write(6, star), s2, s1;
  write(6, star), s2, s2;
  write(6, star), s1, 12;
  write(6, star), s2, 34;
  write(6, star), 56, s1;
  write(6, star), 78, s2;
  write(6, star), "aBcD ", 12, " eFgHi ", 345;
""")
  #
  assert not absd(get("tab_syndrome.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int i = 1;
  write(6, star), i;
  i = 3;
  write(6, star), i;
  i = 5;
  write(6, star), i;
  i = 7;
  write(6, star), i;
  i = 123456;
  write(6, star), i;
""")
  #
  assert not absd(get("ops.f"), -10, """\
  bool la = fem::bool0;
  bool lb = !la;
  bool lc = la && lb;
  lc = la || lb;
  float a = fem::float0;
  float b = fem::float0;
  float c = a + b;
  c = a - b;
  c = a * b;
  c = a / b;
  fem::str<2> sa = "x";
  fem::str<3> sb = "abc";
  fem::str<5> sc = sa + sb;
""")
  #
  assert not absd(get("mod_integers.f"), -10, """\
  common cmn;
  common_write write(cmn);
  write(6, star), fem::mod(13, 5);
""")
  #
  assert not absd(get("do_enddo.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int i = fem::int0;
  FEM_DO(i, 1, 2) {
    write(6, star), i;
  }
  int j = fem::int0;
  FEM_DOSTEP(j, 3, 5, 2) {
    write(6, star), j;
  }
""")
  #
  assert not absd(get("if_endif.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int i = fem::int0;
  if (i == 0) {
    write(6, star), "i is zero.";
  }
""")
  #
  assert not absd(get("if_else_endif.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int i = fem::int0;
  FEM_DO(i, 0, 1) {
    if (i == 0) {
      write(6, star), "i is zero.";
    }
    else {
      write(6, star), "i is not zero.";
    }
  }
""")
  #
  assert not absd(get("if_elseif_else_endif.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int i = fem::int0;
  FEM_DO(i, 0, 2) {
    if (i == 0) {
      write(6, star), "i is zero.";
    }
    else if (i == 1) {
      write(6, star), "i is one.";
    }
    else {
      write(6, star), "i is not zero or one.";
    }
  }
""")
  #
  assert not absd(get("if_write.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int i = fem::int0;
  FEM_DO(i, 3, 4) {
    if (i < 4) {
      write(6, star), "i is less than four.";
    }
    if (i >= 4) {
      write(6, star), "i is greater than or equal to four.";
    }
  }
""")
  #
  assert not absd(get("assign_to_array_elements.f"), -10, """\
  common cmn;
  common_write write(cmn);
  array<float> abc(dimension(2));
  abc.fill(fem::float0);
  abc(1) = 10;
  abc(2) = 20;
  write(6, star), abc(1), abc(2);
""")
  #
  assert not absd(get("parameter_n.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int i = fem::int0;
  static const int n = 2;
  FEM_DO(i, 1, n) {
    write(6, star), i;
  }
""")
  #
  assert not absd(get("do_ix_num.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int ix = fem::int0;
  static const int num = 2;
  FEM_DO(ix, 1, num) {
    write(6, star), ix;
  }
""")
  #
  assert not absd(get("scopes_1.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int ix = fem::int0;
  int ix_sum = fem::int0;
  FEM_DO(ix, 1, 2) {
    ix_sum = ix_sum + ix;
  }
  write(6, star), ix_sum;
""")
  #
  assert not absd(get("scopes_2.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int ix = fem::int0;
  int ix_sum = fem::int0;
  FEM_DO(ix, 1, 3) {
    if (ix == 1) {
      write(6, star), "ix is one.";
    }
    else {
      ix_sum = ix_sum + ix;
    }
    write(6, star), ix_sum;
  }
  int ix_sum_sq = fem::int0;
  FEM_DO(ix, 2, 3) {
    ix_sum_sq = ix_sum_sq + ix * ix;
  }
  write(6, star), ix_sum_sq;
""")
  #
  assert not absd(get("scopes_3.f"), -10, """\
  common cmn;
  common_write write(cmn);
  static const int n_data = 10;
  float d_max = fem::float0;
  int i = fem::int0;
  array<float> data(dimension(n_data));
  data.fill(fem::float0);
  float d = fem::float0;
  if (n_data <= 100) {
    write(6, star), "branch 1.";
  }
  else {
    d_max = 0;
    FEM_DO(i, 1, n_data) {
      d = data(i);
      if (d_max < d) {
        d_max = d;
      }
    }
  }
""")
  #
  assert not absd(get("scopes_4.f"), -10, """\
  common cmn;
  common_write write(cmn);
  static const int n_data = 10;
  float d_max = fem::float0;
  int i = fem::int0;
  array<float> data(dimension(n_data));
  data.fill(fem::float0);
  array<float> d(dimension(1));
  d.fill(fem::float0);
  if (n_data <= 100) {
    write(6, star), "branch 1.";
  }
  else {
    d_max = 0;
    FEM_DO(i, 1, n_data) {
      d(1) = data(i);
      if (d_max < d(1)) {
        d_max = d(1);
      }
    }
  }
""")
  #
  assert not absd(get("arr_float_2.f"), -10, """\
  common cmn;
  common_write write(cmn);
  int i = fem::int0;
  int j = fem::int0;
  array<float, 2> data(dimension(2, 3));
  data.fill(fem::float0);
  FEM_DO(i, 1, 2) {
    FEM_DO(j, 1, 3) {
      data(i, j) = i + 10 * j;
    }
  }
  FEM_DO(j, 1, 3) {
    FEM_DO(i, 1, 2) {
      write(6, star), data(i, j);
    }
  }
""")
  #
  lines = get("subroutine_1.f")
  assert not absd(lines, 4, """\
void
sub1(
  common& cmn)
{
  using namespace fem::major_types;
  common_write write(cmn);
  write(6, star), "output from sub1.";
}

void
sub2(
  common& cmn)
{
  using namespace fem::major_types;
  common_write write(cmn);
  write(6, star), "output from sub2.";
}
""")
  assert not absd(lines, -10, """\
  common cmn;
  common_write write(cmn);
  write(6, star), "first line in prog.";
  sub1(cmn);
  sub1(cmn);
  sub2(cmn);
  sub2(cmn);
  write(6, star), "last line in prog.";
""")
  #
  lines = get("subroutine_2.f")
  assert not absd(lines, 4, expected="""\
void
sub1(
  common& cmn,
  int& i)
{
  using namespace fem::major_types;
  common_write write(cmn);
  i = 3;
  write(6, star), "sub1", i;
  i = 7;
}

void
sub2(
  common& cmn,
  int& i,
  int& j)
{
  using namespace fem::major_types;
  common_write write(cmn);
  j = 4;
  write(6, star), "sub2", i, j;
  i = 8;
  j = 5;
}
""")
  assert not absd(lines, -10, """\
  common cmn;
  common_write write(cmn);
  write(6, star), "first line in prog.";
  int i = fem::int0;
  sub1(cmn, i);
  write(6, star), "prog", i;
  int j = fem::int0;
  sub2(cmn, i, j);
  write(6, star), "prog", i, j;
  write(6, star), "last line in prog.";
""")
  #
  lines = get("subroutine_3.f")
  assert not absd(lines, 4, """\
void
sub1(
  int& num)
{
  using namespace fem::major_types;
  num = 9;
}

void
sub2(
  fem::ref<int> nums,
  int const& nums_size)
{
  using namespace fem::major_types;
  nums(dimension(star));
  int i = fem::int0;
  FEM_DO(i, 1, nums_size) {
    nums(i) = i * 10;
  }
}

void
sub3(
  fem::ref<int> nums,
  int const& nums_size)
{
  using namespace fem::major_types;
  nums(dimension(star));
  int i = fem::int0;
  FEM_DO(i, 1, nums_size) {
    nums(i) = i * 20;
  }
}
""")
  assert not absd(lines, -10, """\
  common cmn;
  common_write write(cmn);
  int num = fem::int0;
  sub1(num);
  write(6, star), "num after sub1:", num;
  int n = 1;
  sub2(num, n);
  write(6, star), "num after sub2", num;
  array<int> nums(dimension(2));
  nums.fill(fem::int0);
  sub1(nums);
  write(6, star), "nums after sub1:", nums(1), nums(2);
  n = 2;
  sub2(nums, n);
  write(6, star), "nums after sub2:", nums(1), nums(2);
  sub3(nums, n);
  write(6, star), "nums after sub3:", nums(1), nums(2);
""")
  #
  assert not absd(get("combine_decl_1.f"), -10, """\
  array<int> vals(dimension(2));
  vals.fill(fem::int0);
  write(6, star), vals(1);
  array<float> abc(dimension(4));
  abc.fill(fem::float0);
  write(6, star), abc(3);
""")
  #
  assert not absd(get("implied_program.f"), -9, """\
void
program_main(
  int argc,
  char const* argv[])
{
  using namespace fem::major_types;
  if (argc != 1) {
    throw std::runtime_error("Unexpected command-line arguments.");
  }
  common cmn;
  common_write write(cmn);
  int num = fem::int0;
  write(6, star), num;
}
""")
  #
  lines = get("implied_trailing_program.f")
  assert not absd(lines, 4, expected="""\
void
sub(
  common& cmn)
{
  using namespace fem::major_types;
  common_write write(cmn);
  write(6, star), "write sub";
}
""")
  assert not absd(lines, -10, """\
  sub(cmn);
""")
  #
  lines = get("common_0.f")
  assert not absd(lines, 1, expected="""\

struct com
{
  int num;

  com() :
    num(fem::int0)
  {}
};

""")
  assert not absd(lines, -10, """\
  common cmn;
  common_write write(cmn);
  write(6, star), cmn.num;
""")
  #
  lines = get("common_1.f")
  assert not absd(lines, 2, expected="""\
struct com
{
  int num;

  com() :
    num(fem::int0)
  {}
};

struct common :
  fem::common,
  com
{};

void
sub(
  common& cmn)
{
  using namespace fem::major_types;
  common_write write(cmn);
  write(6, star), cmn.num;
}
""")
  assert not absd(lines, -10, """\
  common cmn;
  sub(cmn);
  cmn.num = 7;
  sub(cmn);
""")
  #
  lines = get("common_2.f")
  assert not absd(lines, 2, expected="""\
struct com
{
  int num;
  float val;

  com() :
    num(fem::int0),
    val(fem::float0)
  {}
};

struct common :
  fem::common,
  com
{};

void
sub(
  common& cmn)
{
  using namespace fem::major_types;
  common_write write(cmn);
  write(6, star), cmn.num;
  write(6, star), cmn.val;
}
""")
  assert not absd(lines, -10, """\
  common cmn;
  cmn.num = 3;
  cmn.val = 9;
  sub(cmn);
""")
  #
  lines = get("common_3.f")
  assert not absd(lines, 2, expected="""\
struct com
{
  fem::array<int> vals;

  com() :
    vals(fem::dimension(2), fem::fill0)
  {}
};

struct common :
  fem::common,
  com
{};

void
sub(
  common& cmn)
{
  using namespace fem::major_types;
  cmn.vals(1) = cmn.vals(2) + 1;
}
""")
  assert not absd(lines, -10, """\
  common cmn;
  common_write write(cmn);
  cmn.vals(2) = 4;
  sub(cmn);
  write(6, star), cmn.vals(1);
""")
  #
  lines = get("common_4.f")
  assert not absd(lines, 2, expected="""\
struct com
{
  fem::array<int> n;

  com() :
    n(fem::dimension(2), fem::fill0)
  {}
};

struct common :
  fem::common,
  com
{};

void
sub(
  common& cmn,
  int const& num)
{
  using namespace fem::major_types;
  cmn.n(1) = num + 1;
  cmn.n(2) = num + 3;
}
""")
  assert not absd(lines, -10, """\
  common cmn;
  common_write write(cmn);
  write(6, star), cmn.n(1), cmn.n(2);
  sub(cmn, 5);
  write(6, star), cmn.n(2), cmn.n(1);
""")
  #
  for file_name in ["save_0.f", "save_1.f"]:
    lines = get(file_name)
    assert not absd(lines, 1, expected="""\

struct prog_save
{
  int num;

  prog_save() :
    num(fem::int0)
  {}
};

struct common :
  fem::common
{
  prog_save prog_sve;
};

""")
    assert not absd(lines, -10, """\
  common cmn;
  prog_save& sve = cmn.prog_sve;
  common_write write(cmn);
  write(6, star), sve.num;
""")
  #
  lines = get("save_2.f")
  assert not absd(lines, 1, expected="""\

struct sub_save
{
  int num;

  sub_save() :
    num(fem::int0)
  {}
};

struct common :
  fem::common
{
  sub_save sub_sve;
};

void
sub(
  common& cmn)
{
  using namespace fem::major_types;
  sub_save& sve = cmn.sub_sve;
  common_write write(cmn);
  write(6, star), sve.num;
  sve.num = sve.num + 1;
}

""")
  assert not absd(lines, -10, """\
  common cmn;
  sub(cmn);
  sub(cmn);
""")
  #
  lines = get("conv_recipe.f")
  assert not absd(lines, 1, expected="""\

struct abc
{
  float a;
  float b;
  float c;

  abc() :
    a(fem::float0),
    b(fem::float0),
    c(fem::float0)
  {}
};

struct show_resolution_save
{
  fem::data_init_flag data_init;
  float ass;
  float bss;
  float css;
  bool first;

  show_resolution_save() :
    ass(fem::float0),
    bss(fem::float0),
    css(fem::float0),
    first(fem::bool0)
  {}
};

struct common :
  fem::common,
  abc
{
  show_resolution_save show_resolution_sve;
};

void
show_resolution(
  common& cmn,
  int const& h,
  int const& k,
  int const& l)
{
  using namespace fem::major_types;
  show_resolution_save& sve = cmn.show_resolution_sve;
  common_write write(cmn);
  if (sve.data_init()) {
    data((values, true)), sve.first;
  }
  if (sve.first) {
    sve.first = false;
    if (cmn.a <= 0 || cmn.b <= 0 || cmn.c <= 0) {
      write(0, "(1x,a)"), "invalid unit cell constants.";
      FEM_STOP(0);
    }
    sve.ass = 1 / (cmn.a * cmn.a);
    sve.bss = 1 / (cmn.b * cmn.b);
    sve.css = 1 / (cmn.c * cmn.c);
  }
  float dss = h * h * sve.ass + k * k * sve.bss + l * l * sve.css;
  if (dss == 0) {
    write(6, "(3(1x,i3),1x,a)"), h, k, l, "    infinity";
  }
  else {
    write(6, "(3(1x,i3),1x,f12.6)"), h, k, l, fem::sqrt(1 / dss);
  }
}

""")
  assert not absd(lines, -10, """\
  common cmn;
  cmn.a = 11.0f;
  cmn.b = 12.0f;
  cmn.c = 13.0f;
  show_resolution(cmn, 0, 0, 0);
  show_resolution(cmn, 1, 2, 3);
""")
  #
  assert not absd(get("read_star_integer.f"), -10, """\
  common_read read(cmn);
  common_write write(cmn);
  int num = fem::int0;
  read(5, star), num;
  write(6, star), num + 1;
""")
  #
  assert not absd(get("write_implied_do_1.f"), -10, """\
  int i = fem::int0;
  {
    write_loop wloop(cmn, 6, "(2i4)");
    FEM_DO(i, 1, 2) {
      wloop, nums(i);
    }
  }
  int j = fem::int0;
  {
    write_loop wloop(cmn, 6, "(3i4)");
    FEM_DO(i, 1, 2) {
      wloop, nums(i);
      FEM_DO(j, 1, 2) {
        wloop, nums(j);
      }
    }
  }
  {
    write_loop wloop(cmn, 6, "(5i4)");
    FEM_DO(i, 1, 2) {
      wloop, nums(i);
      FEM_DO(j, 1, 2) {
        wloop, nums(j);
      }
      FEM_DO(j, 1, 2) {
        wloop, nums(j);
      }
    }
  }
  int k = fem::int0;
  {
    write_loop wloop(cmn, 6, "(7i4)");
    FEM_DO(i, 1, 2) {
      wloop, nums(i);
      FEM_DO(j, 1, 2) {
        wloop, nums(j);
        FEM_DO(k, 1, 2) {
          wloop, nums(k);
        }
      }
    }
  }
""")
  #
  lines = get("write_extra_parentheses.f")
  assert not absd(lines, -45, """\
int
i1(
  int const& i)
{
  using namespace fem::major_types;
  int function_i1;
  function_i1 = 7 * i;
  return function_i1;
}

int
i2(
  int const& i,
  int const& j)
{
  using namespace fem::major_types;
  int function_i2;
  function_i2 = i * 8 + j;
  return function_i2;
}

int
i3(
  int const& i,
  int const& j,
  int const& k)
{
  using namespace fem::major_types;
  int function_i3;
  function_i3 = i * 9 + j * 29 + k;
  return function_i3;
}
""")
  assert not absd(lines, -10, """\
  common cmn;
  common_write write(cmn);
  int i = 3;
  int j = 4;
  int k = 5;
  int l = 6;
  int m = 7;
  write(6, "(1i5)"), -i * 3;
  write(6, "(1i5)"), -i * 3;
  write(6, "(2i5)"), -i * 3, -j * 4;
  write(6, "(3i5)"), -i * 3, -j * 4, -k * 7;
  write(6, "(3i5)"), -i * 3, -j * 4, -k * 7;
  write(6, "(3i5)"), -i * 3, -j * 4, -k * 7;
  write(6, "(4i5)"), -i * 3, -j * 4, -k * 7, -l * 8;
  write(6, "(4i5)"), -i * 3, -j * 4, -k * 7, -l * 8;
  write(6, "(4i5)"), -i * 3, -j * 4, -k * 7, -l * 8;
  write(6, "(5i5)"), -i * 3, -j * 4, -k * 7, -l * 8, -m * 9;
  write(6, "(4i5)"), i1(-i * 3), -j * 4, -k * 7, -l * 8;
  write(6, "(4i5)"), i1(-i * 3), -j * 4, -k * 7, -l * 8;
  write(6, "(4i5)"), i1(-i * 3), -j * 4, -k * 7, -l * 8;
  write(6, "(4i5)"), i1(-i * 3), -j * 4, -k * 7, -l * 8;
  write(6, "(4i5)"), i1(-i * 3), -j * 4, -k * 7, -l * 8;
  write(6, "(2i5)"), i1(-i * 3), i3(-j * 4, -k * 7, -l * 8);
  write(6, "(2i5)"), i1(-i * 3), i3(-j * 4, -k * 7, -l * 8);
  write(6, "(2i5)"), i1(-i * 3), i3(-j * 4, -k * 7, -l * 8);
""")
  #
  for file_name in [
        "function_two_returns_1.f",
        "function_two_returns_2.f"]:
    lines = get(file_name)
  assert not absd(lines, -24, """\
int
fun(
  int const& i)
{
  using namespace fem::major_types;
  int function_fun;
  if (i < 3) {
    function_fun = i * 4;
    return function_fun;
  }
  function_fun = -i;
  return function_fun;
}
""")
  assert not absd(lines, -10, """\
  common cmn;
  common_write write(cmn);
  write(6, star), fun(2);
  write(6, star), fun(3);
""")
  #
  assert not absd(get("do_label.f"), -10, """\
  int num = fem::int0;
  FEM_DO(num, 1, 2) {
    write(6, star), num;
  }
  int num2 = fem::int0;
  FEM_DO(num, 1, 2) {
    FEM_DO(num2, 3, 4) {
      write(6, star), num, num2;
    }
  }
  FEM_DO(num, 1, 2) {
    FEM_DO(num2, 3, 4) {
      write(6, star), num, num2;
      write(6, star), num * 10, num2 * 10;
    }
  }
""")
  #
  assert not absd(get("label_format.f"), -10, """\
  int num = 102;
  write(6, "(2(i4,1x,2i5,/,'_^'))"), num, num + 1, num + 2, num + 3, num + 4, num + 5;
""")
  #
  assert not absd(get("write_internal_file.f"), -10, """\
  int num = -2;
  fem::str<5> buf = fem::char0;
  write(buf, "(i3)"), num;
  write(6, "('num = (',a,')')"), buf;
""")
  #
  assert not absd(get("open_chain.f"), -10, """\
  int ios = fem::int0;
  cmn.io.open(10, "fable_tmp_661de075");
  cmn.io.close(10);
  cmn.io.open(10, "fable_tmp_661de075")
    .form("formatted")
    .status("unknown");
  cmn.io.close(10);
  try {
    cmn.io.open(10, "fable_tmp_661de075")
      .access("direct")
      .form("unformatted")
      .recl(100)
      .status("new")
      .iostat(ios);
  }
  catch (fem::open_err const&) {
    goto statement_10;
  }
  goto statement_20;
  statement_10:
  write(6, "(a)"), "open err statement";
  statement_20:
  write(6, "(a)"), "Done.";
""")
  #
  assert not absd(get("goto_spaghetti.f"), -10, """\
  int i = fem::int0;
  int j = fem::int0;
  write(6, star), "start";
  goto statement_20;
  statement_10:
  i = 3;
  write(6, star), "stmt 10";
  goto statement_30;
  statement_20:
  write(6, star), "stmt 20";
  i = 2;
  statement_30:
  write(6, star), "stmt 30", i;
  if (i == 2) {
    goto statement_10;
  }
  FEM_DO(j, 1, 2) {
    if (j == 2) {
      goto statement_40;
    }
    write(6, star), "loop j is", j;
    statement_40:;
  }
  write(6, star), "end";
""")
  #
  lines = get("sub_nums_size_capacity.f")
  assert not absd(lines, 4, """\
void
sub(
  common& cmn,
  fem::cref<int> nums,
  int const& nums_size,
  int const& nums_capacity)
{
  using namespace fem::major_types;
  nums(dimension(nums_capacity));
  common_write write(cmn);
  int i = fem::int0;
  FEM_DO(i, 1, nums_size) {
    write(6, "(i3)"), nums(i);
  }
}
""")
  assert not absd(lines, -10, """\
  array<int> nums(dimension(3));
  nums.fill(fem::int0);
  nums(1) = 12;
  nums(2) = 34;
  sub(cmn, nums, 2, 3);
""")
  #
  lines = get("passing_arrays.f")
  assert not absd(lines, 4, """\
void
sub(
  common& cmn,
  int const& num,
  fem::cref<int> nums1,
  fem::cref<int, 2> nums2)
{
  using namespace fem::major_types;
  nums1(dimension(6));
  nums2(dimension(2, 3));
  common_write write(cmn);
  write(6, "(i1)"), num;
  int i = fem::int0;
  FEM_DO(i, 1, 6) {
    write(6, "(i2)"), nums1(i);
  }
  int j = fem::int0;
  FEM_DO(i, 1, 2) {
    FEM_DO(j, 1, 3) {
      write(6, "(i2)"), nums2(i, j);
    }
  }
}
""")
  assert not absd(lines, -10, """\
  array<int> nums(dimension(6));
  nums.fill(fem::int0);
  nums(1) = 3;
  int i = fem::int0;
  FEM_DO(i, 2, 6) {
    nums(i) = nums(i - 1) + i;
  }
  sub(cmn, nums, nums, nums);
""")
  #
  lines = get("unused_args.f")
  assert not absd(lines, 4, """\
void
sub(
  int const& /* num */,
  fem::cref<int> /* nums1 */,
  fem::cref<int, 2> /* nums2 */)
{
  using namespace fem::major_types;
}
""")
  assert not absd(lines, -10, """\
  sub(1, 2, 3);
""")
  #
  lines = get("passing_arrays_2.f")
  assert not absd(lines, 4, """\
void
sub(
  fem::ref<int> nums1,
  fem::ref<int, 2> nums2)
{
  using namespace fem::major_types;
  nums1(dim1(0, 1));
  nums2(dim1(2, 4).dim2(-1, 2));
  nums1(0) = 23;
  nums1(1) = 45;
  nums2(4, -1) = 67;
  int i = fem::int0;
  int j = fem::int0;
  FEM_DO(i, 2, 4) {
    FEM_DO(j, 0, 2) {
      nums2(i, j) = i * 10 + j;
    }
  }
}
""")
  assert not absd(lines, -10, """\
  array<int> nums(dimension(12));
  nums.fill(fem::int0);
  sub(nums, nums);
  int i = fem::int0;
  {
    write_loop wloop(cmn, 6, "(6i3)");
    FEM_DO(i, 1, 12) {
      wloop, nums(i);
    }
  }
""")
  #
  lines = get("array_origin.f")
  assert not absd(lines, 16, """\
  array<int> nums1(dim1(0, 1));
  nums1.fill(fem::int0);
  int k = fem::int0;
  array<int, 2> nums2(dim1(0, 1).dim2(-1, 2));
  nums2.fill(fem::int0);
  int j = fem::int0;
  array<int, 3> nums3(dim1(0, 1).dim2(3).dim3(-1, 2));
""")
  #
  lines = get("power.f")
  assert not absd(lines, -10, """\
  static const float val = fem::pow(1.2f, 3.4f);
  write(6, "(f5.3)"), val;
  float x = 1.2f + fem::pow(3.4f, 5.6f) / 7.8f;
  write(6, "(f5.1)"), x;
  x = fem::pow((1.2f + 3.4f), 5.6f) / 7.8f;
  write(6, "(f5.1)"), x;
  x = fem::pow((1.2f + 3.4f), (5.6f / 7.8f));
  write(6, "(f5.3)"), x;
  x = -fem::pow(1.3f, 2);
  write(6, "(f5.2)"), x;
  x = fem::pow((-1.3f), 2);
  write(6, "(f4.2)"), x;
""")
  #
  lines = get("stop_bare.f")
  assert not absd(lines, -11, """\
    if (i == 2) {
      FEM_STOP(0);
    }
    write(6, "(a,i2)"), "iteration", i;
""")
  #
  lines = get("stop_integer.f")
  assert not absd(lines, -11, """\
    if (i == 2) {
      FEM_STOP(2345);
    }
    write(6, "(a,i2)"), "iteration", i;
""")
  #
  lines = get("stop_string.f")
  assert not absd(lines, -11, """\
    if (i == 2) {
      FEM_STOP("Break");
    }
    write(6, "(a,i2)"), "iteration", i;
""")
  #
  lines = get("passing_strings.f")
  assert not absd(lines, 4, """\
void
sub(
  common& cmn,
  fem::str_cref str)
{
  using namespace fem::major_types;
  common_write write(cmn);
  write(6, "(i1)"), fem::len(str);
  write(6, "(a)"), str;
}
""")
  assert not absd(lines, -10, """\
  fem::str<2> str2 = "Pq";
  fem::str<3> str3 = "rSt";
  sub(cmn, str2);
  sub(cmn, str3);
""")
  #
  lines = get("write_internal_file_2.f")
  assert not absd(lines, -10, """\
  array<int> nums(dimension(2));
  nums.fill(fem::int0);
  nums(1) = -2;
  nums(2) = 3;
  fem::str<8> buf = fem::char0;
  int i = fem::int0;
  {
    write_loop wloop(buf, "(2i3)");
    FEM_DO(i, 1, 2) {
      wloop, nums(i);
    }
  }
  write(6, "('nums = (',a,')')"), buf;
""")
  #
  lines = get("open_write_read.f")
  assert not absd(lines, 21, """\
  cmn.io.open(10, buf)
    .form("formatted");
  int num = fem::int0;
  read(10, "(i6)"), num;
  cmn.io.close(10);
""")
  #
  lines = get("write_internal_file_3.f")
  assert not absd(lines, -10, """\
  int num = -2;
  fem::str<6> buf = "AbCdEf";
  write(buf(2, 4), "(i3)"), num;
  write(6, "('num = (',a,')')"), buf;
""")
  #
  lines = get("open_write_read_2.f")
  assert not absd(lines, -10, """\
  cmn.io.open(10, "fable_tmp_7895777d")
    .form("unformatted")
    .status("unknown");
  write(10), -123;
  cmn.io.close(10);
  cmn.io.open(10, "fable_tmp_7895777d")
    .form("unformatted")
    .status("old");
  int num = fem::int0;
  read(10), num;
  cmn.io.close(10);
  if (num !=  - 123) {
    write(6, "(a)"), "FAILURE int", num;
  }
  else {
    write(6, "(a)"), "OK";
  }
""")
  #
  lines = get("read_err.f")
  assert not absd(lines, -10, """\
  try {
    read(10, "(i1)"), num;
  }
  catch (fem::read_err const&) {
    goto statement_10;
  }
  write(6, "(a)"), "FAILURE: err=10 expected";
  goto statement_20;
  statement_10:
  write(6, "(a)"), "OK";
  statement_20:;
""")
  #
  lines = get("read_end.f")
  assert not absd(lines, -15, """\
  try {
    read(10, "(i1)"), num1, num2;
  }
  catch (fem::read_end const&) {
    goto statement_10;
  }
""")
  #
  lines = get("read_end_err.f")
  assert not absd(lines, -15, """\
  try {
    read(10, "(i1)"), num1, num2;
  }
  catch (fem::read_end const&) {
    goto statement_10;
  }
  catch (fem::read_err const&) {
    goto statement_20;
  }
""")
  #
  lines = get("write_err.f")
  assert not absd(lines, -16, """\
  try {
    write(10, "(i1)"), num;
  }
  catch (fem::write_err const&) {
    goto statement_10;
  }
""")
  #
  lines = get("write_read_end_err_implied_do.f")
  assert not absd(lines, -34, """\
  try {
    write_loop wloop(cmn, 10, "(2i3)");
    FEM_DO(i, 8, 9) {
      wloop, i + 23;
    }
  }
  catch (fem::write_err const&) {
    goto statement_10;
  }
  statement_10:
""")
  assert not absd(lines, -17, """\
  try {
    read_loop rloop(cmn, 10, "(2i3)");
    FEM_DO(i, 1, 2) {
      rloop, nums(i);
    }
  }
  catch (fem::read_end const&) {
    goto statement_20;
  }
  catch (fem::read_err const&) {
    goto statement_30;
  }
  statement_20:
  statement_30:
""")
  #
  lines = get("goto_last_do.f")
  assert not absd(lines, 12, """\
  if (num == 1) {
    num = 2;
    goto statement_10;
  }
  num = 3;
  statement_10:
  FEM_DO(i, 1, num) {
    write(6, "(i2)"), i;
  }
""")
  #
  lines = get("inquire.f")
  assert not absd(lines, 18, """\
  cmn.io.inquire_unit(10)
    .name(str8);
  write(6, "(3a)"), "[", str8, "]";
  cmn.io.inquire_file("fable_tmp_5d70aa2a")
    .opened(lopened);
  write(6, "(a,l1,a)"), "(", lopened, ")";
""")
  assert not absd(lines, 27, """\
  try {
    cmn.io.inquire_file("fable_tmp_5d70aa2a")
      .exist(lexist);
  }
  catch (fem::inquire_err const&) {
    goto statement_10;
  }
  write(6, "(a,l1,a)"), "(", lexist, ")";
  goto statement_20;
  statement_10:
""")
  #
  lines = get("data_type_star.f")
  assert not absd(lines, -10, """\
  fem::logical_star_1 l1 = false;
  fem::integer_star_2 i2 = 4;
  fem::integer_star_4 i4 = 8;
  fem::integer_star_8 i8 = fem::zero<fem::integer_star_8>();
  if (i2 * 2 == i4) {
    i8 = 16;
    if (i4 * 2 == i8) {
      write(6, "(a)"), "OK integers";
      l1 = true;
    }
  }
  if (!l1) {
    write(6, "(a)"), "FAILURE integers";
  }
  fem::real_star_4 r4 = 3.14f;
  fem::real_star_8 r8 = 6.28f;
  if (fem::abs(r4 * 2 - r8) < 1.e-5f) {
    write(6, "(a)"), "OK reals";
  }
  else {
    write(6, "(a)"), "FAILURE reals";
  }
""")
  #
  lines = get("integer_star_2_array.f")
  assert not absd(lines, -10, """\
  array<fem::integer_star_2> nums(dimension(2));
  nums.fill(fem::zero<fem::integer_star_2>());
  nums(1) = 9;
  nums(2) = -6;
  int num_sum = nums(1) + nums(2);
  write(6, "(i1)"), num_sum;
""")
  #
  lines = get("power_2.f")
  assert not absd(lines, -17, """\
  vals(1) = 1.2f;
  vals(2) = fem::pow(vals(1), 2);
  vals(3) = fem::pow(2, vals(2));
  vals(4) = fem::pow(vals(2), vals(3));
""")
  #
  lines = get("subroutine_5.f")
  assert not absd(lines, 4, """\
void
sub(
  common& cmn,
  fem::str_cref str)
{
  using namespace fem::major_types;
  common_write write(cmn);
  if (str(1, 1) == " ") {
    write(6, "(a)"), "str starts with a blank";
  }
  else {
    write(6, "(a)"), "str does not start with a blank";
  }
}
""")
  #
  lines = get("string_compare.f")
  assert not absd(lines, -10, """\
  str2 = " y";
  write(6, "(a,2l1)"), "p", str2 == " y", str2 != " y";
  write(6, "(a,2l1)"), "q", str2 == " y ", str2 != " y ";
  write(6, "(a,2l1)"), "r", str2 == " yz", str2 != " yz";
  write(6, "(a,2l1)"), "s", " y" == str2, " y" != str2;
""")
  #
  lines = get("decl_before_if.f")
  assert not absd(lines, 11, """\
  int num = fem::int0;
  if (num_max > 41) {
    num = 41;
  }
  else {
    num = num_max;
  }
""")
  #
  lines = get("const_analysis_1.f", top_unit_name="prog")
  assert not absd(lines, 4, """\
void
sub1(
  int& num)
{
  using namespace fem::major_types;
  num = 12;
}

void
sub2(
  int& num)
{
  using namespace fem::major_types;
  sub1(num);
}
""")
  #
  lines = get("const_analysis_2.f", top_unit_name="prog")
  assert not absd(lines, 4, """\
void
sub1(
  common& cmn,
  int& num)
{
  using namespace fem::major_types;
  common_read read(cmn);
  read(5, "(i2)"), num;
}

void
sub2(
  common& cmn,
  int& num)
{
  using namespace fem::major_types;
  sub1(cmn, num);
}
""")
  #
  lines = get("double_literal.f")
  assert not absd(lines, -10, """\
  write(6, star), 1.2e2, 3.4e2f, 5.6e2;
""")
  #
  lines = get("read_implied_do_1.f")
  assert not absd(lines, 4, """\
void
sub(
  common& cmn,
  fem::ref<int> nums)
{
  using namespace fem::major_types;
  nums(dimension(star));
  common_read read(cmn);
  int i = fem::int0;
  {
    read_loop rloop(cmn, 5, "(2i3)");
    FEM_DO(i, 1, 2) {
      rloop, nums(i);
    }
  }
}
""")
  #
  lines = get("dim_with_parameter.f")
  assert not absd(lines, -14, """\
  static const int num = 2;
  array<int> nums1(dimension(num));
  nums1.fill(fem::int0);
  nums1(1) = 12;
  nums1(2) = 34;
  array<int> nums2(dimension(num));
""")

def exercise_syntax_error(verbose):
  t_dir = libtbx.env.under_dist(
    module_name="fable", path="test/syntax_error", test=op.isdir)
  from fable.read import Error
  def fail(file_name):
    if (verbose):
      print "exercise_syntax_error:", file_name
    cout.process(file_names=[op.join(t_dir, file_name)])
  try:
    fail("bad_open_err_label.f")
  except Error, e:
    assert str(e).startswith("Invalid statement label:")
    assert str(e).endswith("""\
  |      open(1, file=name, err=1.3)|
--------------------------------^""")
  else: raise Exception_expected
  try:
    fail("power_no_base.f")
  except Error, e:
    assert str(e).startswith("Syntax error:")
    assert str(e).endswith("""\
  |      x = **3.4|
-------------^""")
  else: raise Exception_expected
  try:
    fail("power_no_exponent.f")
  except Error, e:
    assert str(e).startswith("Syntax error:")
    assert str(e).endswith("""\
  |      x = 1.2**|
----------------^""")
  else: raise Exception_expected

def exercise_semantic_error(verbose):
  t_dir = libtbx.env.under_dist(
    module_name="fable", path="test/semantic_error", test=op.isdir)
  from fable import SemanticError
  def fail(file_name):
    if (verbose):
      print "exercise_semantic_error:", file_name
    cout.process(file_names=[op.join(t_dir, file_name)])
  try:
    fail("assignment_to_parameter.f")
  except SemanticError, e:
    assert str(e).startswith("Assignment to PARAMETER n:")
    assert str(e).endswith("""\
  |      n = 1|
---------^""")
  else: raise Exception_expected
  try:
    fail("inquire_no_unit_no_file.f")
  except SemanticError, e:
    assert str(e).startswith("Missing UNIT or FILE in INQUIRE statement:")
    assert str(e).endswith("""\
  |      inquire(exist=lexist)|
---------^""")
  else: raise Exception_expected
  try:
    fail("inquire_both_unit_and_file.f")
  except SemanticError, e:
    assert str(e).startswith(
      "Conflicting UNIT vs. FILE in INQUIRE statement"
      " (exactly one is needed):")
    assert str(e).endswith("""\
  |      inquire(10, file='fable_tmp')|
---------^""")
  else: raise Exception_expected
  try:
    fail("recursion_in_declaration.f")
  except SemanticError, e:
    assert str(e).startswith("Recursion in declaration:")
    assert str(e).endswith("""\
  |      dimension nums(nums)|
------------------------^""")
  else: raise Exception_expected

def exercise_unsupported(verbose):
  t_dir = libtbx.env.under_dist(
    module_name="fable", path="test/unsupported", test=op.isdir)
  def get(file_name):
    if (verbose):
      print "exercise_unsupported:", file_name
    return "\n".join(
      cout.process(file_names=[op.join(t_dir, file_name)])).splitlines()
  #
  assert not absd(get("goto_into_loop.f"), -10, """\
  int i = fem::int0;
  FEM_DO(i, 1, 2) {
    statement_10:
    write(6, star), i;
  }
  goto statement_10;
""")

def run(args):
  assert args in [[], ["--verbose"]]
  verbose = (len(args) != 0)
  exercise_simple(verbose=verbose)
  exercise_syntax_error(verbose=verbose)
  exercise_semantic_error(verbose=verbose)
  exercise_unsupported(verbose=verbose)
  print "OK"

if (__name__ == "__main__"):
  import sys
  run(args=sys.argv[1:])
